// Copyright Coffee Stain Studios. All Rights Reserved.

#pragma once

#include "CoreMinimal.h"
#include "FGBuildableFactory.h"
#include "FGBuildableRailroadTrack.h"
#include "FGBuildableTrainPlatform.generated.h"


UENUM( BlueprintType )
enum class ETrainPlatformDockingStatus : uint8
{
	ETPDS_None						UMETA( DisplayName = "None" ),
	ETPDS_WaitingToStart			UMETA( DisplayName = "WaitingToStart" ),
	ETPDS_Loading					UMETA( DisplayName = "Loading" ),
	ETPDS_Unloading					UMETA( DisplayName = "Unloading" ), 
	ETPDS_WaitingForTransfer		UMETA( DisplayName = "WaitingTransfer" ),
	ETPDS_Complete					UMETA( DisplayName = "Complete" ),
	ETPDS_WaitForTransferCondition	UMETA( DisplayName = "WaitForCondition" ),
	ETPDS_IdleWaitForTime			UMETA( DisplayName = "Idle" )
};

/**
 *	Base class for any train platform (station, cargo loading, empty etc. )
 */
UCLASS( )
class FACTORYGAME_API AFGBuildableTrainPlatform : public AFGBuildableFactory
{
	GENERATED_BODY()
public:
	AFGBuildableTrainPlatform();

	// Begin Actor interface
	virtual void BeginPlay() override;
	virtual void GetLifetimeReplicatedProps( TArray<FLifetimeProperty>& OutLifetimeProps ) const override;
	// End Actor interface

	// Begin save interface
	virtual void PreSaveGame_Implementation( int32 saveVersion, int32 gameVersion ) override;
	virtual void PostLoadGame_Implementation( int32 saveVersion, int32 gameVersion ) override;
	// End save interface

	// Begin IFGDismantlableInterface
	virtual bool CanDismantle_Implementation() const override;
	virtual void GetChildDismantleActors_Implementation(TArray<AActor*>& out_ChildDismantleActors) const override;
	virtual void Dismantle_Implementation() override;
	virtual bool SupportsDismantleDisqualifiers_Implementation() const override { return true; }
	// End IFGDismantlableInterface

	//~ Begin IFGUseableInterface
	virtual bool IsUseable_Implementation() const override;
	//~ End IFGUseableInterface

	//@todo-trains Get rid of the dependency on track position in platforms.
	FRailroadTrackPosition GetTrackPosition() const;
	int32 GetTrackGraphID() const;

	/** Get the platform that precedes this platform
	*	@param direction - 0 or 1 
	*	@info this can be null if the platform in the specified direction does not exist
	*/
	AFGBuildableTrainPlatform* GetConnectedPlatformInDirectionOf( uint8 direction );

	/** When a locomotive docks it will call this on relevant children in the direction of its output. It is up to the platform to decide how to act */
	virtual void NotifyTrainDocked( class AFGRailroadVehicle* railroadVehicle, class AFGBuildableRailroadStation* initiatedByStation );

	/** Is this platform reversed? */
	FORCEINLINE bool IsOrientationReversed() const { return mIsOrientationReversed; }

	UFUNCTION( BlueprintPure, Category = "FactoryGame|Railroad|Platform" )
	ETrainPlatformDockingStatus GetDockingStatus() const { return mPlatformDockingStatus; }

	/** True if the player has Cancelled a dock from the FGTrain */
	UFUNCTION( BlueprintPure, Category = "FactoryGame|Railroad|Platform" )
	bool IsDockingCancelled() { return mDockWasCancelled; }

	/** Timer Sequence Function to progress and handle all logic for docking
	 *	When a platform is docked, a timer handle is created that updates the dock sequence until its completion that calls this function each step
	 */
	virtual	void UpdateDockingSequence();

	/** Cancel the current dock (if there is one) */
	virtual void CancelDockingSequence();

protected:
	/** his needs to happed from the hologram and from begin play */
	virtual void SetupRailroadTrack();

	/** Call to clear all docking related properties, overrides should always call super */
	virtual void FinishDockingSequence();

	UFUNCTION()
	virtual void OnRep_RailroadTrack();
	UFUNCTION()
	virtual void OnRep_UpdateDockingStatus();
	UFUNCTION()
	virtual void OnRep_DockedRailroadVehicle();

private:
	/** Used by the hologram to configure this platform. */
	void ReverseConnectionDirections();
	void AssignRailroadTrack( class AFGBuildableRailroadTrack* track );

protected:
	UPROPERTY( SaveGame, ReplicatedUsing = OnRep_RailroadTrack )
	class AFGBuildableRailroadTrack* mRailroadTrack;

	/** Each platform should have exactly 2 FGTrainPlatformConnectionComponents (Not a uproperty as it just holds refs to the below connections)
	*	They are individual components so that they appear correctly in the blueprint editor. If anyone sees this and knows a way to make containers
	*	of components be editable when inherited @me (dylan)
	*/
	UPROPERTY()
	TArray< class UFGTrainPlatformConnection* > mPlatformConnections;
	
	UPROPERTY( EditAnywhere )
	class UFGTrainPlatformConnection* mPlatformConnection0;

	UPROPERTY( EditAnywhere )
	class UFGTrainPlatformConnection* mPlatformConnection1;

	UPROPERTY( ReplicatedUsing = OnRep_DockedRailroadVehicle, SaveGame )
	class AFGRailroadVehicle* mDockedRailroadVehicle;

	/** Stores a reference to the station that initiated a docking sequence. Used to notify the station that we have completed */
	UPROPERTY( SaveGame )
	class AFGBuildableRailroadStation* mStationDockingMaster;

	// Is this platform reversed from its attached station?
	UPROPERTY( SaveGame )
	bool mIsOrientationReversed;

	// Where are we in the docking status. Updated by the mDockingSequenceTimer which is set by NotifyTrainDocked
	UPROPERTY( ReplicatedUsing = OnRep_UpdateDockingStatus )
	ETrainPlatformDockingStatus mPlatformDockingStatus;

	// The save state value to write for the stations docking status. It is modified in presave to force a docking reset if saved part way through.
	// However since it also drives the anim blueprint undesired anim starts are triggered if we save / modify the actually used mPlatformDockingStatus
	UPROPERTY( SaveGame )
	ETrainPlatformDockingStatus mSavedDockingStatus;

	// Timer handle that is set and retriggered over the course of a docking sequence
	UPROPERTY()
	FTimerHandle mDockingSequenceTimerHandle;

	// Timer handle to track when updates to the docking sequence are in the Idle state
	UPROPERTY()
	FTimerHandle mIdleUpdateTimerHandle;

	UPROPERTY( Replicated )
	bool mDockWasCancelled;

private:
	friend class AFGTrainPlatformHologram;
};
